@@ -1,6 +1,6 @@
 /* 
    neon test suite
-   Copyright (C) 2002-2009, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2025, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#include <errno.h>
 
 #include "ne_request.h"
 #include "ne_socket.h"
@@ -66,8 +67,7 @@ static ne_ssl_client_cert *def_cli_cert;
 static char *nul_cn_fn;
 
 static int check_dname(const ne_ssl_dname *dn, const char *expected,
-                       const char *which)
-    ne_attribute((nonnull));
+                       const char *which);
 
 static int check_cert_dnames(const ne_ssl_certificate *cert,
                              const char *subject, const char *issuer)
@@ -76,7 +76,8 @@ static int check_cert_dnames(const ne_ssl_certificate *cert,
 /* Arguments for running the SSL server */
 struct ssl_server_args {
     char *cert; /* the server cert to present. */
-    const char *response; /* the response to send. */
+    const char *key; /* server key filename */
+    const char *response; /* the HTTP response message to send. */
     int numreqs; /* number of request/responses to handle over the SSL connection. */
 
     /* client cert handling: */
@@ -92,9 +93,8 @@ struct ssl_server_args {
     } session;
     int count; /* internal use. */
 
-    int use_ssl2; /* force use of SSLv2 only */
-
-    const char *key;
+    /* minimum, maximum version. */
+    enum ne_ssl_protocol minver, maxver;
 };
 
 /* default response string if args->response is NULL */
@@ -110,8 +110,7 @@ static int ssl_server(ne_socket *sock, void *userdata)
     static ne_ssl_context *ctx = NULL;
 
     if (ctx == NULL) {
-        ctx = ne_ssl_context_create(args->use_ssl2 ? NE_SSL_CTX_SERVERv2
-                                    : NE_SSL_CTX_SERVER);
+        ctx = ne_ssl_context_create(NE_SSL_CTX_SERVER);
     }
 
     ONV(ctx == NULL, ("could not create SSL context"));
@@ -163,6 +162,18 @@ static int ssl_server(ne_socket *sock, void *userdata)
         
     } while (--args->numreqs > 0);
 
+    if (args->minver || args->maxver) {
+        unsigned int vers = ne_sock_getproto(sock);
+        const char *name = ne_ssl_proto_name(vers);
+
+        NE_DEBUG(NE_DBG_HTTP, "ssl: TLS protocol version (%d): [%s]\n", 
+                 vers, name);
+
+        ONV((args->minver && vers < args->minver)
+            || (args->maxver && vers > args->maxver),
+            ("SSL protocol version %d (%s) outside range (%d, %d)",
+             vers, name, args->minver, args->maxver));
+    }
 
     if (args->cache) {
         unsigned char sessid[128];
@@ -174,7 +185,7 @@ static int ssl_server(ne_socket *sock, void *userdata)
 #ifdef NE_DEBUGGING
         {
             char *b64 = ne_base64(sessid, len);
-            NE_DEBUG(NE_DBG_SSL, "Session id retrieved (%d): [%s]\n", 
+            NE_DEBUG(NE_DBG_SSL, "ssl: Session id retrieved (%d): [%s]\n", 
                      args->count, b64);
             ne_free(b64);
         }
@@ -191,19 +202,13 @@ static int ssl_server(ne_socket *sock, void *userdata)
                 || memcmp(args->session.id, sessid, len));
         }
     }	
+
+    ret = ne_sock_shutdown(sock, NE_SOCK_SEND);
+    NE_DEBUG(NE_DBG_SSL, "ssl: Shutdown received %d\n", ret);
     
     return 0;
 }
 
-/* serve_ssl wrapper which ignores server failure and always succeeds */
-static int fail_serve(ne_socket *sock, void *ud)
-{
-    struct ssl_server_args args = {0};
-    args.cert = ud;
-    ssl_server(sock, &args);
-    return OK;
-}
-
 #define DEFSESS  (ne_session_create("https", "localhost", 7777))
 
 static int make_ssl_session_port(ne_session **sess,
@@ -470,28 +475,6 @@ static int simple(void)
     return accept_signed_cert(SERVER_CERT);
 }
 
-#if 0 /* No longer works for modern SSL libraries, rightly so. */
-/* Test for SSL operation when server uses SSLv2 */
-static int simple_sslv2(void)
-{
-    ne_session *sess = ne_session_create("https", "localhost", 7777);
-    struct ssl_server_args args = {SERVER_CERT, 0};
-
-    args.use_ssl2 = 1;
-    ne_set_session_flag(sess, NE_SESSFLAG_SSLv2, 1);
-
-    if (ne_get_session_flag(sess, NE_SESSFLAG_SSLv2) != 1) {
-        t_context("no SSLv2 support in SSL library");
-        ne_session_destroy(sess);
-        return SKIP;
-    }
-
-    CALL(any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL));
-    ne_session_destroy(sess);
-    return OK;
-}
-#endif
-
 /* Test read-til-EOF behaviour with SSL. */
 static int simple_eof(void)
 {
@@ -593,11 +576,6 @@ static int ipaddr_altname(void)
     return accept_signed_cert_for_hostname("altname5.cert", "127.0.0.1");
 }
 
-static int uri_altname(void)
-{
-    return accept_signed_cert_for_hostname("altname7.cert", "localhost");
-}
-
 /* test that the *most specific* commonName attribute is used. */
 static int multi_commonName(void)
 {
@@ -769,6 +747,7 @@ static int get_failures(void *userdata, int fs, const ne_ssl_certificate *c)
 {
     int *out = userdata;
     *out = fs;
+    NE_DEBUG(NE_DBG_SSL, "test: fail_ssl_request verify callback - %d\n", fs);
     return -1;
 }
 
@@ -789,42 +768,28 @@ static int fail_ssl_request_with_error2(char *cert, char *key, char *cacert,
                                         const char *msg, int failures,
                                         const char *errstr)
 {
-    ne_session *sess = ne_session_create("https", host, 7777);
+    ne_session *sess;
     int gotf = 0, ret;
     struct ssl_server_args args = {0};
     ne_sock_addr *addr = NULL;
     const ne_inet_addr **list = NULL;
 
-    if (realhost) {
-        size_t n;
-        const ne_inet_addr *ia;
-
-        addr = ne_addr_resolve(realhost, 0);
-
-        ONV(ne_addr_result(addr),
-            ("fake hostname lookup failed for %s", realhost));
-
-        NE_DEBUG(NE_DBG_SSL, "ssl: Using fake hostname '%s'\n", realhost);
-
-        for (n = 0, ia = ne_addr_first(addr); ia; ia = ne_addr_next(addr))
-            n++;
-
-        NE_DEBUG(NE_DBG_SSL, "ssl: Address count '%lu'\n", n);
-
-        list = ne_calloc(n * sizeof(*list));
-
-        for (n = 0, ia = ne_addr_first(addr); ia; ia = ne_addr_next(addr))
-            list[n++] = ia;
-        
-        ne_set_addrlist(sess, list, n);
-    }
-
     args.cert = cert;
     args.key = key;
     args.fail_silently = 1;
     
-    ret = any_ssl_request(sess, ssl_server, &args, cacert,
-			  get_failures, &gotf);
+    CALL(fakeproxied_session_server(&sess, "https", host, 7777, ssl_server, &args));
+
+    if (cacert) {
+        CALL(load_and_trust_cert(sess, cacert));
+    }
+
+    ne_ssl_set_verify(sess, get_failures, &gotf);
+
+    ret = any_request(sess, "/expect-to-fail");
+
+    NE_DEBUG(NE_DBG_SSL, "test: fail_ssl_request - request code %d, error: %s\n",
+             ret, ne_get_error(sess));
 
     ONV(gotf == 0,
 	("no error in verification callback; request rv %d error string: %s",
@@ -955,18 +920,22 @@ static int fail_self_signed(void)
  * commonName (and no alt names either). */
 static int fail_missing_CN(void)
 {
-    ne_session *sess = DEFSESS;
+    struct ssl_server_args args = {0};
+    ne_session *sess;
+    int ret;
 
-    ONN("accepted server cert with missing commonName",
-        any_ssl_request(sess, fail_serve, "missingcn.cert", SERVER_CERT,
-                        NULL, NULL) == NE_OK);
-    
-    ONV(strstr(ne_get_error(sess), "missing commonName") == NULL,
-        ("unexpected session error `%s'", ne_get_error(sess)));
+    args.cert = "missingcn.cert";
 
-    ne_session_destroy(sess);
-    return OK;
-}                            
+    CALL(make_ssl_session(&sess, "localhost", ssl_server, &args));
+
+    ret = any_request(sess, "/fail-missing-cn");
+    ONN("request did not fail", ret != NE_ERROR);
+
+    ONV(strstr(ne_get_error(sess), "missing commonName attribute") == NULL,
+        ("error string unexpected: %s", ne_get_error(sess)));
+
+    return destroy_and_wait(sess);
+}
 
 /* test for a bad ipAddress altname */
 static int fail_bad_ipaltname(void)
@@ -983,12 +952,6 @@ static int fail_host_ipaltname(void)
                             "bad IP altname cert", NE_SSL_IDMISMATCH);
 }
 
-static int fail_bad_urialtname(void)
-{
-    return fail_ssl_request("altname8.cert", CA_CERT, "localhost",
-                            "bad URI altname cert", NE_SSL_IDMISMATCH);
-}
-
 static int fail_wildcard(void)
 {
     return fail_ssl_request("altname9.cert", CA_CERT, "localhost",
@@ -1015,7 +978,6 @@ static int fail_ca_notyetvalid(void)
                             "issuer ca not yet valid", NE_SSL_BADCHAIN);
 }
 
-#if 0
 /* Test that the SSL session is cached across connections. */
 static int session_cache(void)
 {
@@ -1028,6 +990,10 @@ static int session_cache(void)
     CALL(multi_session_server(&sess, "https", "localhost",
                               2, ssl_server, &args));
 
+    /* This currently fails under OpenSSL with TLSv1.3. */
+    ne_ssl_set_protovers(sess, NE_SSL_PROTO_UNSPEC,
+                         NE_SSL_PROTO_TLS_1_2);
+
     ne_ssl_trust_cert(sess, def_ca_cert);
 
     ONREQ(any_request(sess, "/req1"));
@@ -1035,7 +1001,6 @@ static int session_cache(void)
 
     return destroy_and_wait(sess);
 }
-#endif
 
 /* Callback for client_cert_provider; takes a c. cert as userdata and
  * registers it. */
@@ -1292,27 +1257,42 @@ static int proxy_tunnel(void)
     return OK;
 }
 
-#define RESP_0LENGTH "HTTP/1.1 200 OK\r\n" "Content-Length: 0\r\n" "\r\n"
+struct tunnel_args {
+    int iteration;
+    const char *first_response; /* first CONNECT response. */
+    const char *second_response; /* second CONNECT response. */
+    struct ssl_server_args *args;
+};
 
-/* a tricky test which requires spawning a second server process in
- * time for a new connection after a 407. */
-static int apt_post_send(ne_request *req, void *ud, const ne_status *st)
+/* Server which acts as a proxy accepting a CONNECT request. */
+static int serve_auth_tunnel(ne_socket *sock, void *ud)
 {
-    int *code = ud;
-    if (st->code == *code) {
-        struct ssl_server_args args = {SERVER_CERT, NULL};
-    
-        if (*code == 407) args.numreqs = 2;
-        args.response = RESP_0LENGTH;
+    struct tunnel_args *args = ud;
 
-        NE_DEBUG(NE_DBG_HTTP, "Got challenge, awaiting server...\n");
-        CALL(await_server());
-        NE_DEBUG(NE_DBG_HTTP, "Spawning proper tunnel server...\n");
-        /* serve *two* 200 OK responses. */
-        CALL(spawn_server(7777, serve_tunnel, &args));
-        NE_DEBUG(NE_DBG_HTTP, "Spawned.\n");
+    /* check for a server auth function */
+    want_header = "Authorization";
+    got_header = tunnel_header;
+    got_server_auth = 0;
+
+    CALL(discard_request(sock));
+
+    if (got_server_auth) {
+        SEND_STRING(sock, "HTTP/1.1 500 Leaked Server Auth Creds\r\n"
+                    "Content-Length: 0\r\n" "Server: serve_tunnel\r\n\r\n");
+        return 0;
     }
-    return OK;
+    
+    if (args->iteration++ == 0) {
+        /* give the plaintext tunnel reply, acting as the proxy */
+
+        SEND_STRING(sock, args->first_response);
+
+        return OK;
+    }
+
+    SEND_STRING(sock, args->second_response);
+
+    return ssl_server(sock, args->args);
 }
 
 static int apt_creds(void *userdata, const char *realm, int attempt,
@@ -1328,23 +1308,27 @@ static int apt_creds(void *userdata, const char *realm, int attempt,
  * 0.24.0. */
 static int auth_proxy_tunnel(void)
 {
-    ne_session *sess = ne_session_create("https", "localhost", 443);
-    int ret, code = 407;
+    struct ssl_server_args args = {SERVER_CERT, NULL};
+    struct tunnel_args tunnel;
+    ne_session *sess;
+
+    tunnel.first_response =
+        "HTTP/1.0 407 I WANT MORE BISCUITS\r\n"
+        "Server: auth_proxy_tunnel\r\n"
+        "Proxy-Authenticate: Basic realm=\"bigbluesea\"\r\n"
+        "Connection: close\r\n" "\r\n";
+    tunnel.second_response = "HTTP/1.1 200 OK\r\n"
+        "Server: auth_proxy_tunnel r2\r\n\r\n";
+    tunnel.iteration = 0;
+    tunnel.args = &args;
+
+    CALL(proxied_multi_session_server(2, &sess, "https", "localhost", 443,
+                                      serve_auth_tunnel, &tunnel));
     
-    ne_session_proxy(sess, "localhost", 7777);
-    ne_hook_post_send(sess, apt_post_send, &code);
     ne_set_proxy_auth(sess, apt_creds, NULL);
     ne_ssl_trust_cert(sess, def_ca_cert);
     
-    CALL(spawn_server(7777, single_serve_string,
-                      "HTTP/1.0 407 I WANT MORE BISCUITS\r\n"
-                      "Proxy-Authenticate: Basic realm=\"bigbluesea\"\r\n"
-                      "Connection: close\r\n" "\r\n"));
-    
-    /* run two requests over the tunnel. */
-    ret = any_2xx_request(sess, "/foobar");
-    if (!ret) ret = any_2xx_request(sess, "/foobar2");
-    CALL(ret);
+    CALL(any_2xx_request(sess, "/foobar"));
 
     return destroy_and_wait(sess);
 }
@@ -1353,20 +1337,27 @@ static int auth_proxy_tunnel(void)
  * proxy in a CONNECT request. */
 static int auth_tunnel_creds(void)
 {
-    ne_session *sess = ne_session_create("https", "localhost", 443);
-    int code = 401;
-    struct ssl_server_args args = {SERVER_CERT, 0};
+    struct ssl_server_args args = {SERVER_CERT, NULL};
+    struct tunnel_args tunnel;
+    ne_session *sess;
+
+    args.response = "HTTP/1.1 401 I want a Shrubbery\r\n"
+        "WWW-Authenticate: Basic realm=\"bigredocean\"\r\n"
+        "Server: auth_tunnel_creds\r\n" "Content-Length: 0\r\n" "\r\n"
+        ""
+        "HTTP/1.1 200 OK\r\n\r\n";
+
+    tunnel.second_response = "HTTP/1.1 200 OK\r\n\r\n";
+    tunnel.args = &args;
+    tunnel.iteration = 1;
+
+    CALL(proxied_multi_session_server(2, &sess, "https", "localhost", 443,
+                                      serve_auth_tunnel, &tunnel));
     
-    ne_session_proxy(sess, "localhost", 7777);
-    ne_hook_post_send(sess, apt_post_send, &code);
     ne_set_server_auth(sess, apt_creds, NULL);
     ne_ssl_trust_cert(sess, def_ca_cert);
     
-    args.response = "HTTP/1.1 401 I want a Shrubbery\r\n"
-        "WWW-Authenticate: Basic realm=\"bigredocean\"\r\n"
-        "Server: Python\r\n" "Content-Length: 0\r\n" "\r\n";
-    
-    CALL(spawn_server(7777, serve_tunnel, &args));
+    CALL(any_2xx_request(sess, "/foobar"));
     CALL(any_2xx_request(sess, "/foobar"));
 
     return destroy_and_wait(sess);
@@ -1374,25 +1365,34 @@ static int auth_tunnel_creds(void)
 
 static int auth_tunnel_fail(void)
 {
-    ne_session *sess = ne_session_create("https", "localhost", 443);
+    ne_session *sess;
     int ret;
 
-    CALL(spawn_server(7777, single_serve_string,
-                      "HTTP/1.1 407 Nyaaaaah\r\n"
-                      "Proxy-Authenticate: GaBoogle\r\n"
-                      "Connection: close\r\n"
-                      "\r\n"));
-    
-    ne_session_proxy(sess, "localhost", 7777);
+    CALL(proxied_session_server(&sess, "https", "localhost", 7777,
+                                single_serve_string,
+                                "HTTP/1.1 407 Nyaaaaah\r\n"
+                                "Proxy-Authenticate: GaBoogle\r\n"
+                                "Content-Length: 0\r\n"
+                                "\r\n"
+                                "HTTP/1.1 200 OK\r\n"
+                                "Content-Length: 0\r\n"
+                                "\r\n"));
 
     ne_set_proxy_auth(sess, apt_creds, NULL);
-     
-    ret = any_request(sess, "/bar");
-    ONV(ret != NE_PROXYAUTH, ("bad error code for tunnel failure: %d", ret));
 
+    /* First request: should fail and the auth failure should
+     * propagate back to the session error string. */
+    ret = any_request(sess, "/foo");
+    ONV(ret != NE_PROXYAUTH, ("bad error code for tunnel failure: %d", ret));
     ONV(strstr(ne_get_error(sess), "GaBoogle") == NULL,
         ("bad error string for tunnel failure: %s", ne_get_error(sess)));
 
+    /* Second request must fail to connect, the connection should have
+     * been closed. */
+    ret = any_request(sess, "/bar");
+    ONV(ret != NE_CONNECT,
+        ("second attempt should fail to connect, got: %d", ret));
+
     return destroy_and_wait(sess);
 }
 
@@ -1502,7 +1502,6 @@ static int cert_identities(void)
         { "altname2.cert", "nohost.example.com" },
         { "altname4.cert", "localhost" },
         { "ca4.pem", "fourth.example.com" },
-        { "altname8.cert", "http://nohost.example.com/" },
         { NULL, NULL }
     };
     int n;
@@ -1823,24 +1822,36 @@ static int nonssl_trust(void)
     ne_session *sess = ne_session_create("http", "www.example.com", 80);
     
     ne_ssl_trust_cert(sess, def_ca_cert);
+    ne_ssl_trust_default_ca(sess);
     
     ne_session_destroy(sess);
 
     return OK;
 }
 
+#ifdef HAVE_PAKCHOIS
+#define PINMAX (2)
+struct pindata {
+    const char *password[PINMAX];
+    ne_buffer *trace;
+};
+
 /* PIN password provider callback. */
 static int pkcs11_pin(void *userdata, int attempt,
                       const char *slot_descr, const char *token_label,
                       unsigned int flags, char *pin)
 {
-    char *sekrit = userdata;
+    struct pindata *data = userdata;
 
-    NE_DEBUG(NE_DBG_SSL, "pkcs11: slot = [%s], token = [%s]\n",
+    NE_DEBUG(NE_DBG_SSL, "pk11: slot = [%s], token = [%s]\n",
              slot_descr, token_label);
 
-    if (attempt == 0) {
-        strcpy(pin, sekrit);
+    ne_buffer_snprintf(data->trace, 200, "pin(%d,%s,%s,%u)\n",
+                       attempt, slot_descr?slot_descr:"[none]",
+                       token_label?token_label:"[none]", flags);
+
+    if (attempt < PINMAX && data->password[attempt]) {
+        strcpy(pin, data->password[attempt]);
         return 0;
     }
     else {
@@ -1848,11 +1859,15 @@ static int pkcs11_pin(void *userdata, int attempt,
     }
 }
 
+#define SLOT_NSS "NSS User Private Key and Certificate Services"
+#define TOKEN_NSS "NSS Certificate DB"
+
 static int nss_pkcs11_test(const char *dbname)
 {
     ne_session *sess = DEFSESS;
     struct ssl_server_args args = {SERVER_CERT, NULL};
     ne_ssl_pkcs11_provider *prov;
+    struct pindata pindata;
     int ret;
 
     args.require_cc = 1;
@@ -1872,7 +1887,11 @@ static int nss_pkcs11_test(const char *dbname)
         return SKIP;
     }
 
-    ne_ssl_pkcs11_provider_pin(prov, pkcs11_pin, "foobar");
+    pindata.password[0] = "notfoobar";
+    pindata.password[1] = "foobar";
+    pindata.trace = ne_buffer_create();
+
+    ne_ssl_pkcs11_provider_pin(prov, pkcs11_pin, &pindata);
     ne_ssl_set_pkcs11_provider(sess, prov);
 
     ret = any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL);
@@ -1880,6 +1899,11 @@ static int nss_pkcs11_test(const char *dbname)
     ne_session_destroy(sess);
     ne_ssl_pkcs11_provider_destroy(prov);
 
+    if (ret == OK)
+        ONCMPN("pin(0," SLOT_NSS "," TOKEN_NSS ",0)\n"
+               "pin(1," SLOT_NSS "," TOKEN_NSS ",0)\n", pindata.trace->data, "pin callback", "data");
+    ne_buffer_destroy(pindata.trace);
+
     return ret;
 }
 
@@ -1892,6 +1916,87 @@ static int pkcs11_dsa(void)
 {
     return nss_pkcs11_test("nssdb-dsa");
 }
+#endif
+
+static int protovers(void)
+{
+    ne_session *sess = DEFSESS;
+    struct ssl_server_args args = {SERVER_CERT, NULL};
+
+    args.minver = NE_SSL_PROTO_TLS_1_2;
+    args.maxver = NE_SSL_PROTO_TLS_1_2;
+
+    ONV(ne_ssl_set_protovers(sess, args.minver, args.maxver),
+        ("setting TLS protocol version failed: %s", ne_get_error(sess)));
+
+    CALL(any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL));
+
+    ne_session_destroy(sess);
+    return OK;
+}
+
+static int notifier(void)
+{
+    ne_session *sess = DEFSESS;
+    struct ssl_server_args args = {SERVER_CERT, NULL};
+    ne_buffer *buf = ne_buffer_create();
+
+    args.minver = NE_SSL_PROTO_TLS_1_2;
+    args.maxver = NE_SSL_PROTO_TLS_1_2;
+
+    ONV(ne_ssl_set_protovers(sess, args.minver, args.maxver),
+        ("setting TLS protocol version failed: %s", ne_get_error(sess)));
+
+    ne_set_notifier(sess, sess_notifier, buf);
+
+    CALL(any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL));
+
+    ONV(strstr(buf->data, "-handshake(TLSv1.2, TLS_") == NULL
+        && strstr(buf->data, "-handshake(TLSv1.2, [none]") == NULL,
+        ("missing handshake from notifier: %s", buf->data));
+
+    ne_session_destroy(sess);
+    ne_buffer_destroy(buf);
+
+    return OK;
+}
+
+#define TEST_URI "./enclient.pem"
+
+static int clicert_uri(void)
+{
+    ne_session *sess;
+    struct ssl_server_args args = {SERVER_CERT, NULL};
+    ne_ssl_client_cert *cc;
+    int ret;
+
+    args.require_cc = 1;
+
+    cc = ne_ssl_clicert_fromuri(TEST_URI, 0);
+    if (!cc && errno == ENOTSUP) {
+        t_context("client certificate URI support not available");
+        return SKIP;
+    }
+    ONV(!cc, ("could not load client certificate URI %s: %s",
+              TEST_URI, strerror(errno)));
+
+    ONN("cc not in encrypted state", !ne_ssl_clicert_encrypted(cc));
+    ONN("successful decrypt with bad password",
+        ne_ssl_clicert_decrypt(cc, "not-the-password") == 0);
+    ONN("unsuccessful decrypt with good password",
+        ne_ssl_clicert_decrypt(cc, P12_PASSPHRASE) != 0);
+
+    sess = DEFSESS;
+    ne_ssl_set_clicert(sess, cc);
+
+    ret = any_ssl_request(sess, ssl_server, &args, CA_CERT, NULL, NULL);
+
+    ne_session_destroy(sess);
+    ne_ssl_clicert_free(cc);
+
+    return ret;
+}
+
 
 /* TODO: code paths still to test in cert verification:
  * - server cert changes between connections: Mozilla gives
@@ -1928,9 +2033,6 @@ ne_test tests[] = {
     T(clicert_import),
 
     T(simple),
-#if 0
-    T(simple_sslv2),
-#endif
     T(simple_eof),
     T(empty_truncated_eof),
     T(fail_not_ssl),
@@ -1958,7 +2060,6 @@ ne_test tests[] = {
     T(two_subject_altname2),
     T(notdns_altname),
     T(ipaddr_altname),
-    T(uri_altname),
 
     T(multi_commonName),
     T(commonName_first),
@@ -1971,7 +2072,6 @@ ne_test tests[] = {
     T(fail_missing_CN),
     T(fail_host_ipaltname),
     T(fail_bad_ipaltname),
-    T(fail_bad_urialtname),
     T(fail_wildcard),
     T(fail_wildcard_ip),
     T(fail_ca_notyetvalid),
@@ -1985,9 +2085,7 @@ ne_test tests[] = {
     T(fail_nul_san),
 #endif
 
-#if 0
     T(session_cache),
-#endif
 
     T(fail_tunnel),
     T(proxy_tunnel),
@@ -1995,10 +2093,16 @@ ne_test tests[] = {
     T(auth_tunnel_creds),
     T(auth_tunnel_fail),
 
+    T(protovers),
+
     T(nonssl_trust),
 
+#ifdef HAVE_PAKCHOIS
     T(pkcs11),
     T_XFAIL(pkcs11_dsa), /* unclear why this fails currently. */
+#endif
+    T(notifier),
+    T(clicert_uri),
 
     T(NULL) 
 };
